"""Decorators and helpers to check access to communities."""

from functools import wraps
from typing import Any, Callable, Dict, Optional, Union

from flask import g
from flask_login import current_user
from werkzeug.exceptions import Forbidden
from werkzeug.wrappers.response import Response

from abilian.core.models.subjects import User
from abilian.sbe.apps.communities.presenters import CommunityPresenter
from abilian.services import get_service
from abilian.services.security import MANAGE


def require_admin(func: Callable) -> Callable:
    @wraps(func)
    def decorated_view(*args: Any, **kwargs: Any) -> Union[str, Response]:
        security = get_service("security")
        is_admin = security.has_role(current_user, "admin")
        if not is_admin:
            raise Forbidden()
        return func(*args, **kwargs)

    return decorated_view


def require_manage(func: Callable) -> Callable:
    @wraps(func)
    def decorated_view(*args: Any, **kwargs: Any) -> Response:
        community = g.community
        if community and community.has_permission(current_user, MANAGE):
            return func(*args, **kwargs)
        security = get_service("security")
        is_admin = security.has_role(current_user, "admin")
        if not is_admin:
            raise Forbidden()
        return func(*args, **kwargs)

    return decorated_view


def require_access(func):
    @wraps(func)
    def decorated_view(*args, **kwargs):
        check_access()
        return func(*args, **kwargs)

    return decorated_view


def check_access(
    community: CommunityPresenter = None, user: Optional[User] = None
) -> None:
    if not has_access(community, user):
        raise Forbidden()


def has_access(
    community: CommunityPresenter = None, user: Optional[User] = None
) -> bool:
    if not user:
        user = current_user
    if user.is_anonymous:
        return False

    security = get_service("security")
    is_admin = security.has_role(user, "admin")
    if is_admin:
        return True

    if not community:
        community = getattr(g, "community", None)

    if community is not None:
        return community.get_role(user) is not None

    return False


def is_manager(
    context: Optional[Dict[str, Any]] = None, user: Optional[User] = None
) -> bool:
    security = get_service("security")

    if not user:
        user = current_user
    if user.is_anonymous:
        return False

    if context:
        community = context.get("object").community  # type: ignore
    else:
        community = g.community

    if community.has_permission(user, MANAGE) or user == community.creator:
        return True

    if security.has_role(user, "admin"):
        return True

    return False



from flask import url_for
from flask.ctx import RequestContext
from flask.testing import FlaskClient

from abilian.core.sqlalchemy import SQLAlchemy
from abilian.sbe.app import Application
from abilian.sbe.apps.communities.models import Community
from abilian.services.security import Admin
from abilian.services.security.service import SecurityService
from abilian.testing.util import client_login


def test_index(
    community1: Community,
    app: Application,
    db: SQLAlchemy,
    client: FlaskClient,
    req_ctx: RequestContext,
) -> None:
    security_service: SecurityService = app.services["security"]
    security_service.start()

    user = community1.test_user
    with client_login(client, user):
        response = client.get(url_for("communities.index"))
        assert response.status_code == 200


def test_community_home(
    community1: Community,
    community2: Community,
    app: Application,
    client: FlaskClient,
    req_ctx: RequestContext,
) -> None:
    security_service: SecurityService = app.services["security"]
    security_service.start()

    url = app.default_view.url_for(community1)

    user1 = community1.test_user

    with client_login(client, user1):
        response = client.get(url)
        assert response.status_code == 302
        expected_url = url_for(
            "wall.index", community_id=community1.slug, _external=True
        )
        assert response.location == expected_url

    user2 = community2.test_user
    with client_login(client, user2):
        response = client.get(url)
        assert response.status_code == 403


def test_new(
    community1: Community,
    app: Application,
    client: FlaskClient,
    db: SQLAlchemy,
    req_ctx: RequestContext,
) -> None:
    security_service: SecurityService = app.services["security"]
    # security_service.use_cache = False
    security_service.start()

    user = community1.test_user

    with client_login(client, user):
        response = client.get(url_for("communities.new"))
        assert response.status_code == 403

    security_service.grant_role(user, Admin)
    db.session.flush()

    with client_login(client, user):
        response = client.get(url_for("communities.new"))
        assert response.status_code == 200


def test_community_settings(
    app: Application,
    client: FlaskClient,
    community1: Community,
    req_ctx: RequestContext,
) -> None:
    security_service: SecurityService = app.services["security"]
    security_service.start()

    url = url_for("communities.settings", community_id=community1.slug)
    user = community1.test_user

    with client_login(client, user):
        response = client.get(url)
        assert response.status_code == 403

        app.services["security"].grant_role(user, Admin)
        response = client.get(url)
        assert response.status_code == 200

        data = {
            "__action": "edit",
            "name": "edited community",
            "description": "my community",
            "linked_group": "",
            "type": "participative",
        }
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200
        assert "edited community" in response.get_data(as_text=True)


def test_members(
    app: Application,
    client: FlaskClient,
    db: SQLAlchemy,
    community1: Community,
    community2: Community,
    req_ctx: RequestContext,
) -> None:
    security_service: SecurityService = app.services["security"]
    security_service.start()

    user1 = community1.test_user
    user2 = community2.test_user

    with client_login(client, user1):
        url = url_for("communities.members", community_id=community1.slug)
        response = client.get(url)
        assert response.status_code == 200

        # test add user
        data = {"action": "add-user-role", "user": user2.id}
        response = client.post(url, data=data)
        assert response.status_code == 403

        security_service.grant_role(user1, Admin)

        data = {"action": "add-user-role", "user": user2.id, "role": "member"}
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200

        membership = [m for m in community1.memberships if m.user == user2][0]
        assert membership.role == "member"

        data["action"] = "set-user-role"
        data["role"] = "manager"
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200

        db.session.expire(membership)
        assert membership.role == "manager"

        # Community.query.session is not self.db.session, but web app
        # session.
        community = Community.query.get(community1.id)
        assert user2 in community.members

        # test delete
        data = {
            "action": "delete",
            "user": user2.id,
            "membership": [m.id for m in community1.memberships if m.user == user2][0],
        }
        response = client.post(url, data=data, follow_redirects=True)
        assert response.status_code == 200
        assert user2 not in community.members



import logging
import time
from datetime import datetime
from pathlib import Path
from typing import Any, List, Optional, Union

import sqlalchemy as sa
from blinker import ANY
from flask import current_app
from sqlalchemy import (
    Boolean,
    DateTime,
    ForeignKey,
    Integer,
    String,
    Unicode,
    UniqueConstraint,
    and_,
)
from sqlalchemy.event import listens_for
from sqlalchemy.orm import backref, relation, relationship
from sqlalchemy.orm.attributes import OP_APPEND, OP_REMOVE, Event
from sqlalchemy.sql.schema import Column
from werkzeug.local import LocalProxy

from abilian.core.entities import Entity, EntityMeta
from abilian.core.extensions import db
from abilian.core.models import NOT_AUDITABLE, SEARCHABLE
from abilian.core.models.blob import Blob
from abilian.core.models.subjects import Group, User
from abilian.i18n import _l
from abilian.sbe.apps.documents.models import Folder
from abilian.sbe.apps.documents.repository import repository
from abilian.services.indexing import indexable_role
from abilian.services.security import READ, WRITE, Admin
from abilian.services.security import Manager as MANAGER
from abilian.services.security import Permission
from abilian.services.security import Reader as READER
from abilian.services.security import Role, RoleType
from abilian.services.security import Writer as WRITER
from abilian.services.security import security

from . import signals

logger = logging.getLogger(__name__)

MEMBER = Role("member", label=_l("role_member"), assignable=False)
VALID_ROLES = frozenset([READER, WRITER, MANAGER, MEMBER])


class Membership(db.Model):
    """Represents the membership of someone in a community."""

    __tablename__ = "community_membership"

    id = Column(Integer, primary_key=True)

    user_id = Column(ForeignKey("user.id"), index=True, nullable=False)
    user = relationship(
        User, lazy="joined", backref=backref("communautes_membership", lazy="select")
    )

    community_id = Column(ForeignKey("community.id"), index=True, nullable=False)
    community = relationship("Community", lazy="joined")

    role = Column(RoleType())  # should be either 'member' or 'manager'

    __table_args__ = (UniqueConstraint("user_id", "community_id"),)

    def __repr__(self):
        return "<Membership user={}, community={}, role={}>".format(
            repr(self.user), repr(self.community), str(self.role)
        ).encode("utf-8")


def community_content(cls: type) -> Any:
    """Class decorator to mark models considered as community content.

    This is required for proper indexation.
    """
    cls.is_community_content = True

    def community_slug(self: Any) -> str:
        return self.community and self.community.slug

    cls.community_slug = property(community_slug)

    index_to = getattr(cls, "__index_to__", ())
    index_to += (("community_slug", ("community_slug",)),)
    cls.__index_to__ = index_to

    def _indexable_roles_and_users(self: Any) -> str:
        if not self.community:
            return []
        return indexable_roles_and_users(self.community)

    cls._indexable_roles_and_users = property(_indexable_roles_and_users)
    return cls


def indexable_roles_and_users(community: "Community") -> str:
    """Mixin to use to replace Entity._indexable_roles_and_users.

    Will be removed when communities are upgraded to use standard role
    based access (by setting permissions and using security service).
    """
    # TODO: remove
    return " ".join(indexable_role(user) for user in community.members)


class Community(Entity):
    """Ad-hoc objects that hold properties about a community."""

    __index_to__ = (("id", ("id", "community_id")),)

    is_community_content = True

    # : A public description.
    description = Column(Unicode(500), default="", nullable=False, info=SEARCHABLE)

    #: An image or logo for this community.
    image_id = Column(ForeignKey(Blob.id), index=True)
    image = relationship(Blob, lazy="joined")

    #: The root folder for this community.
    folder_id = Column(
        Integer,
        ForeignKey(Folder.id, use_alter=True, name="fk_community_root_folder"),
        unique=True,
    )
    folder = relation(
        Folder,
        single_parent=True,  # required for delete-orphan
        primaryjoin=(folder_id == Folder.id),
        cascade="all, delete-orphan",
        backref=backref("_community", lazy="select", uselist=False),
    )

    #: The group this community is linked to, if any. Memberships are then
    #: reflected
    group_id = Column(ForeignKey(Group.id), nullable=True, unique=True)
    group = relation(Group, foreign_keys=group_id, lazy="select")

    #: Memberships for this community.
    memberships = relationship(Membership, cascade="all, delete-orphan")

    #: direct access to :class:`User` members
    members = relationship(
        User,
        secondary=Membership.__table__,
        viewonly=True,
        backref=backref("communities", lazy="select", viewonly=True),
    )

    #: Number of members in this community.
    membership_count = Column(Integer, default=0, nullable=False, info=NOT_AUDITABLE)

    #: Number of documents in this community.
    document_count = Column(Integer, default=0, nullable=False, info=NOT_AUDITABLE)

    #: Last time something happened in this community
    last_active_at = Column(
        DateTime, default=datetime.utcnow, nullable=False, info=NOT_AUDITABLE
    )

    # Various features for this community:

    #: True if this community has a document management space
    has_documents = Column(Boolean, nullable=False, default=True)

    #: True if this community has a wiki
    has_wiki = Column(Boolean, nullable=False, default=True)

    #: True if this community has a forum
    has_forum = Column(Boolean, nullable=False, default=True)

    #: One of 'participative' or 'informative
    type = Column(String(20), nullable=False, default="informative")

    #: One of 'secret', 'public', 'open' (currently not used)
    visibility = Column(String(20), nullable=False, default="secret")

    #: Used for segmenting communities. Not used currently.
    category = Column(String(20), nullable=False, default="")

    #: True if regular members can send stuff by email to all members
    # Not used anymore.
    members_can_send_by_email = Column(Boolean, nullable=False, default=False)

    def __init__(self, **kw):
        self.has_documents = True
        self.membership_count = 0
        self.document_count = 0
        self.members_can_send_by_email = False
        Entity.__init__(self, **kw)
        if self.has_documents and not self.folder:
            # FIXME: this should be done in documents by using signals
            name = self.name
            if not name:
                # during creation, we may have to provide a temporary name for
                # subfolder, we don't want empty names on folders since they must be
                # unique among siblings
                name = f"{self.__class__.__name__}_{str(self.id)}-{time.asctime()}"
            self.folder = repository.root_folder.create_subfolder(name)
            # if not self.group:
            #   self.group = Group(name=self.name)

        if not self.image:
            fn = Path(__file__).parent / "views" / "data" / "community.png"
            self.image = Blob(fn.open("rb").read())

    @property
    def has_calendar(self) -> bool:
        config = current_app.config
        return bool(config.get("ENABLE_CALENDAR"))

    def rename(self, name: str) -> None:
        self.name = name
        if self.folder:
            # FIXME: use signals
            self.folder.name = name

    def get_memberships(self, role: Optional[str] = None) -> List[Membership]:
        M = Membership
        memberships = M.query.filter(M.community_id == self.id)
        if role:
            memberships = memberships.filter(M.role == role)
        return memberships.all()

    def set_membership(self, user: User, role: Union[str, Role]) -> None:
        """Add a member with the given role, or set the role of an existing
        member."""
        assert isinstance(user, User)
        role = Role(role)

        if role not in VALID_ROLES:
            raise ValueError(f"Invalid role: {role}")

        session = sa.orm.object_session(self) or db.session()
        is_new = True
        M = Membership
        membership = (
            session.query(M)
            .filter(and_(M.user_id == user.id, M.community_id == self.id))
            .first()
        )

        if not membership:
            membership = Membership(community=self, user=user, role=role)
            session.add(membership)
            self.membership_count += 1
        else:
            is_new = False
            membership.role = role

        signals.membership_set.send(self, membership=membership, is_new=is_new)

    def remove_membership(self, user: User) -> None:
        M = Membership
        membership = M.query.filter(
            and_(M.user_id == user.id, M.community_id == self.id)
        ).first()
        if not membership:
            raise KeyError(f"User {user} is not a member of community {self}")

        db.session.delete(membership)
        self.membership_count -= 1
        signals.membership_removed.send(self, membership=membership)

    def update_roles_on_folder(self) -> None:
        if self.folder:
            self.ungrant_all_roles_on_folder()
            for membership in self.memberships:
                user = membership.user
                role = membership.role
                if role == MANAGER:
                    security.grant_role(user, MANAGER, self.folder)
                else:
                    if self.type == "participative":
                        security.grant_role(user, WRITER, self.folder)
                    else:
                        security.grant_role(user, READER, self.folder)

    def ungrant_all_roles_on_folder(self) -> None:
        if self.folder:
            role_assignments = security.get_role_assignements(self.folder)
            for principal, role in role_assignments:
                security.ungrant_role(principal, role, self.folder)

    def get_role(self, user: Union[User, LocalProxy]) -> Optional[Role]:
        """Returns the given user's role in this community."""
        M = Membership
        membership = (
            db.session()
            .query(M.role)
            .filter(and_(M.community_id == self.id, M.user_id == user.id))
            .first()
        )

        return membership.role if membership else None

    def has_member(self, user):
        return self.get_role(user) is not None

    def has_permission(self, user: LocalProxy, permission: Permission) -> bool:
        if not isinstance(permission, Permission):
            assert isinstance(permission, str)
            permission = Permission(permission)

        if user.has_role(Admin):
            return True
        role = self.get_role(user)
        if role == MANAGER:
            return True
        if role == MEMBER and permission in (READ, WRITE):
            return True
        return False

    def touch(self) -> None:
        self.last_active_at = datetime.utcnow()

    @property
    def _indexable_roles_and_users(self) -> str:
        return indexable_roles_and_users(self)


def CommunityIdColumn() -> Column:
    return Column(
        ForeignKey(Community.id),
        nullable=False,
        info=SEARCHABLE | {"index_to": (("community_id", ("community_id",)),)},
    )


# Handlers to keep community/group members in sync
#
_PROCESSED_ATTR = "__sbe_community_group_sync_processed__"


@signals.membership_set.connect_via(ANY)
def _membership_added(sender: Community, membership: Membership, is_new: bool) -> None:
    if not is_new:
        return

    if getattr(membership.user, _PROCESSED_ATTR, False) is OP_APPEND:
        return

    if sender.group and membership.user not in sender.group.members:
        logger.debug(
            "_membership_added(%r, %r, %r) user: %r",
            sender,
            membership,
            is_new,
            membership.user,
        )
        setattr(membership.user, _PROCESSED_ATTR, OP_APPEND)
        sender.group.members.add(membership.user)


@signals.membership_removed.connect_via(ANY)
def membership_removed(sender: Community, membership: Membership) -> None:
    if getattr(membership.user, _PROCESSED_ATTR, False) is OP_REMOVE:
        return

    if sender.group and membership.user in sender.group.members:
        logger.debug(
            "_membership_removed(%r, %r) user: %r", sender, membership, membership.user
        )
        setattr(membership.user, _PROCESSED_ATTR, OP_REMOVE)
        sender.group.members.discard(membership.user)


@listens_for(Community.members, "append")
@listens_for(Community.members, "remove")
def _on_member_change(community, user, initiator):
    group = community.group
    if not group:
        return

    logger.debug("_on_member_change(%r, %r, op=%r)", community, user, initiator.op)

    if getattr(user, _PROCESSED_ATTR, False) is initiator.op:
        return

    setattr(user, _PROCESSED_ATTR, initiator.op)

    if initiator.op is OP_APPEND:
        if user not in group.members:
            group.members.add(user)

    elif initiator.op is OP_REMOVE:
        if user in group.members:
            group.members.discard(user)


@listens_for(Community.group, "set", active_history=True)
def _on_linked_group_change(
    community: Community, value: Any, oldvalue: Any, initiator: Event
) -> None:
    if value == oldvalue:
        return

    logger.debug("_on_linked_group_change(%r, %r, %r)", community, value, oldvalue)

    if oldvalue is not None and oldvalue.members:
        logger.debug(
            "_on_linked_group_change(%r, %r, %r): oldvalue clear()",
            community,
            value,
            oldvalue,
        )
        oldvalue.members.clear()

    members = set(community.members)
    if value is not None and value.members != members:
        logger.debug(
            "_on_linked_group_change(%r, %r, %r): set value.members",
            community,
            value,
            oldvalue,
        )
        value.members = members


def _safe_get_community(group: Group) -> Optional[Community]:
    session = sa.orm.object_session(group)
    if not session:
        return None

    with session.no_autoflush:
        try:
            return (
                session.query(Community)
                .filter(Community.group == group)
                .options(
                    sa.orm.joinedload(Community.group),
                    sa.orm.joinedload(Community.members),
                )
                .one()
            )
        except sa.orm.exc.NoResultFound:
            return None


@listens_for(Group.members, "append")
@listens_for(Group.members, "remove")
def _on_group_member_change(group: Group, user: User, initiator: Event) -> None:
    community = _safe_get_community(group)

    if not community:
        return

    op = initiator.op
    if getattr(user, _PROCESSED_ATTR, False) is op:
        return

    is_present = user in community.members
    setattr(user, _PROCESSED_ATTR, op)
    logger.debug(
        "_on_group_member_change(%r, %r, op=%r) community: %r",
        group,
        user,
        initiator.op,
        community,
    )

    if (op is OP_APPEND and is_present) or (op is OP_REMOVE and not is_present):
        return

    if op is OP_APPEND:
        community.set_membership(user, MEMBER)

    elif op is OP_REMOVE:
        community.remove_membership(user)


@listens_for(Group.members, "set", active_history=True)
def _on_group_members_replace(group, value, oldvalue, initiator):
    if value == oldvalue:
        return

    community = _safe_get_community(group)
    if not community:
        return

    members = set(community.members)
    logger.debug(
        "_on_group_members_replace(%r, %r, %r) community: %r",
        group,
        value,
        oldvalue,
        community,
    )

    for u in members - value:
        if getattr(u, _PROCESSED_ATTR, False) is OP_REMOVE:
            continue
        setattr(u, _PROCESSED_ATTR, OP_REMOVE)
        community.remove_membership(u)

    for u in value - members:
        if getattr(u, _PROCESSED_ATTR, False) is OP_APPEND:
            continue
        setattr(u, _PROCESSED_ATTR, OP_APPEND)
        community.set_membership(u, MEMBER)



"""Communities module."""
from flask import Flask


def register_plugin(app: Flask) -> None:
    # Used for side-effect
    from . import events  # noqa
    from . import search
    from .views import communities

    app.register_blueprint(communities)

    search.init_app(app)



from pathlib import Path

from flask import (
    current_app,
    flash,
    jsonify,
    make_response,
    redirect,
    render_template,
    request,
    url_for,
)
from flask_babel import gettext as _
from flask_login import current_user
from sqlalchemy import func
from werkzeug.exceptions import InternalServerError, NotFound

from abilian.core.extensions import db
from abilian.core.models.subjects import Group, User
from abilian.services import get_service
from abilian.services.image import CROP, resize
from abilian.services.security import security
from abilian.web import csrf
from abilian.web.views import default_view

from .social import social

DEFAULT_GROUP_MUGSHOT = (
    (Path(__file__).parent / ".." / ".." / ".." / "static" / "images" / "frog.jpg")
    .open("rb")
    .read()
)


@social.route("/groups/")
def groups() -> str:
    tab = request.args.get("tab", "all_groups")
    if tab == "all_groups":
        groups = Group.query.order_by(Group.name).all()
        if not security.has_role(current_user, "admin"):
            groups = [
                group
                for group in groups
                if group.public or current_user in group.members
            ]
    else:
        groups = current_user.groups
        groups.sort(key=lambda x: x.name)

    return render_template("social/groups.html", groups=groups)


def is_admin(group):
    security = get_service("security")
    is_admin = current_user in group.admins
    if not is_admin and "security" in current_app.extensions:
        is_admin = security.has_role(current_user, "admin")

    return is_admin


@social.route("/groups/<int:group_id>")
@default_view(social, Group, "group_id")
def group_home(group_id):
    group = Group.query.get(group_id)
    ctx = {"group": group, "is_admin": is_admin(group)}
    return render_template("social/group.html", **ctx)


@social.route("/groups/<int:group_id>/json")
def group_json(group_id):
    members = Group.query.get(group_id).members
    q = request.args.get("q", "").lower()
    if q:
        members = [
            u
            for u in members
            if any(
                term.startswith(q)
                for name in (u.first_name.lower(), u.last_name.lower())
                for term in name.split()
            )
        ]

    result = {"results": [{"id": obj.id, "text": obj.name} for obj in members]}
    return jsonify(result)


@social.route("/groups/<int:group_id>", methods=["POST"])
@csrf.protect
def group_post(group_id):
    group = Group.query.get(group_id)
    action = request.form.get("action")
    return_url = request.form.get("return_url")
    user_id = request.form.get("user", "").strip()

    membership_actions = frozenset(("add", "remove", "add-admin", "remove-admin"))

    if action not in {"join", "leave"} | membership_actions:
        raise ValueError(f"Unknown action: {action}")

    if action not in ("join", "leave"):
        assert is_admin(group)

    if action == "join":
        current_user.join(group)
    elif action == "leave":
        current_user.leave(group)

    elif action in membership_actions:
        try:
            user_id = int(user_id)
        except ValueError:
            flash(_("Error: No user selected"), "error")
            return redirect(url_for(".group_home", group_id=group_id))

        user = User.query.get(user_id)

        if action == "add":
            user.join(group)
        elif action == "remove":
            user.leave(group)
        elif action == "add-admin":
            if user not in group.admins:
                group.admins.append(user)
        elif action == "remove-admin":
            if user in group.admins:
                group.admins.remove(user)

    db.session.commit()

    if not return_url:
        return_url = url_for(".group_home", group_id=group_id)

    # TODO: security check
    return redirect(return_url)


@social.route("/groups/new")
def groups_new():
    # TODO later
    return

    # e = Env()
    # e.form = GroupForm()
    # return render_template("social/groups-new.html", **e)


@social.route("/groups/new", methods=["POST"])
def groups_new_post():
    # TODO later
    return

    # form = GroupForm()
    #
    # if form.validate():
    #     group = Group()
    #     form.populate_obj(group)
    #     db.session.add(group)
    #     db.session.commit()
    #     flash(_(u"Your new group has been created"), category='info')
    #     return redirect(url_for('.group_home', group_id=group.id))
    #
    # else:
    #     e = Env()
    #     e.form = form
    #     return render_template("social/groups-new.html", **e)


@social.route("/groups/<int:group_id>/mugshot")
def group_mugshot(group_id):
    # TODO: duplicated code (with user_mugshot). Extract common method.
    size = int(request.args.get("s", 55))
    if size > 500:
        raise ValueError(f"Error, size = {size:d}")
    group = Group.query.get(group_id)

    if not group:
        raise NotFound()

    if group.photo:
        data = group.photo
    else:
        data = DEFAULT_GROUP_MUGSHOT

    if size:
        data = resize(data, size, size, mode=CROP)

    response = make_response(data)
    response.headers["content-type"] = "image/jpeg"
    response.headers.add("Cache-Control", "public, max-age=600")
    return response


#
# Ad-hoc JSON endpoints, used by select boxes
#
@social.route("/groups/json")
def groups_json():
    q = request.args.get("q").replace("%", " ").lower()

    if not q or len(q) < 2:
        raise InternalServerError()

    query = Group.query
    # query = query.filter(func.lower(Group.name).like(q + "%"))
    query = query.order_by(func.lower(Group.name))
    all = query.all()

    result = {"results": [{"id": obj.id, "text": obj.name} for obj in all]}
    return jsonify(result)



"""Forms for the Wiki module."""
from typing import Any, Optional

from wtforms import HiddenField, StringField, TextAreaField, ValidationError
from wtforms.validators import data_required

from abilian.i18n import _, _l
from abilian.web.forms import Form
from abilian.web.forms.filters import strip
from abilian.web.forms.validators import flaghidden
from abilian.web.forms.widgets import TextArea

from .util import page_exists


def clean_up(src: str) -> str:
    """Form filter."""
    src = src.replace("\r", "")
    return src


def int_or_none(val: Any) -> Optional[int]:
    try:
        return int(val)
    except (TypeError, ValueError):
        return None


class WikiPageForm(Form):
    title = StringField(
        label=_l("Title"), filters=(strip,), validators=[data_required()]
    )
    body_src = TextAreaField(
        label=_l("Body"),
        filters=(strip, clean_up),
        validators=[data_required()],
        widget=TextArea(rows=10, resizeable="vertical"),
    )

    message = StringField(label=_l("Commit message"))
    page_id = HiddenField(filters=(int_or_none,), validators=[flaghidden()])
    last_revision_id = HiddenField(filters=(int_or_none,), validators=[flaghidden()])

    def validate_title(self, field: StringField) -> None:
        title = field.data
        if title != field.object_data and page_exists(title):
            raise ValidationError(
                _("A page with this name already exists. Please use another name.")
            )

    def validate_last_revision_id(self, field: HiddenField) -> None:
        val = field.data
        current = field.object_data

        if val is None or current is None:
            return

        if val != current:
            raise ValidationError(_("this page has been edited since"))


# Not used yet
class CommentForm(Form):
    message = TextAreaField(label=_l("Message"), validators=[data_required()])
    page_id = HiddenField()



from datetime import datetime, timedelta
from typing import Any, Dict, Optional, Union

import dateutil.parser
from flask import current_app
from flask_login import current_user

from abilian.core.models.subjects import User
from abilian.core.util import utcnow

DEFAULT_LIFETIME = 3600


class Lock:
    """Represent a lock on a document."""

    def __init__(
        self,
        user_id: int,
        user: str,
        date: Union[datetime, str],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.user_id = user_id
        self.user = user
        if not isinstance(date, datetime):
            try:
                date = dateutil.parser.parse(date)
            except Exception as e:
                raise ValueError(f"Error parsing date: {date!r}") from e

        self.date = date

    @staticmethod
    def new() -> "Lock":
        return Lock(current_user.id, str(current_user), utcnow())

    def as_dict(self) -> Dict[str, Any]:
        """Return a dict suitable for serialization to JSON."""
        return {
            "user_id": self.user_id,
            "user": self.user,
            "date": self.date.isoformat(),
        }

    @staticmethod
    def from_dict(d: Dict[str, Any]) -> "Lock":
        """Deserialize from a `dict` created by :meth:`as_dict`."""
        return Lock(**d)

    @property
    def lifetime(self) -> int:
        return current_app.config.get("SBE_LOCK_LIFETIME", DEFAULT_LIFETIME)

    @property
    def expired(self) -> bool:
        return (utcnow() - self.date) > timedelta(seconds=self.lifetime)

    def is_owner(self, user: Optional[User] = None) -> bool:
        if user is None:
            user = current_user

        return self.user_id == user.id



import difflib
from pathlib import Path
from typing import Any, Dict, Optional, Tuple, Union
from urllib.parse import quote

import sqlalchemy as sa
from flask import (
    current_app,
    flash,
    g,
    make_response,
    redirect,
    render_template,
    request,
)
from flask_login import current_user
from markdown import markdown
from markupsafe import Markup
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import NotFound
from werkzeug.wrappers import Response
from whoosh.searching import Hit

from abilian.core.extensions import db
from abilian.core.signals import activity
from abilian.core.util import unwrap
from abilian.i18n import _, _l, _n
from abilian.sbe.apps.communities.blueprint import Blueprint
from abilian.sbe.apps.communities.common import object_viewers
from abilian.sbe.apps.communities.models import Community
from abilian.sbe.apps.communities.presenters import CommunityPresenter
from abilian.sbe.apps.communities.views import default_view_kw as community_dv_kw
from abilian.services.security.models import Permission
from abilian.services.viewtracker import viewtracker
from abilian.web import csrf
from abilian.web.action import Endpoint, actions
from abilian.web.nav import BreadcrumbItem
from abilian.web.util import url_for
from abilian.web.views import ObjectCreate, ObjectEdit, ObjectView, default_view

from .forms import WikiPageForm
from .models import WikiPage, WikiPageAttachment, WikiPageRevision

wiki = Blueprint("wiki", __name__, url_prefix="/wiki", template_folder="templates")
route = wiki.route


@wiki.url_value_preprocessor
def init_wiki_values(endpoint: str, values: Dict[Any, Any]) -> None:
    g.current_tab = "wiki"

    endpoint = Endpoint("wiki.index", community_id=g.community.slug)
    g.breadcrumb.append(BreadcrumbItem(label=_l("Wiki"), url=endpoint))

    title = request.args.get("title", "").strip()
    if title and title != "Home":
        url = Endpoint("wiki.page", community_id=g.community.slug, title=title)
        g.breadcrumb.append(BreadcrumbItem(label=title, url=url))


@route("/")
def index() -> Response:
    return redirect(url_for(".page", title="Home", community_id=g.community.slug))


def wiki_page_default_view_kw(kw, obj, obj_type, obj_id, **kwargs):
    kw = community_dv_kw(kw, obj, obj_type, obj_id, **kwargs)
    title = kwargs.get("title")
    if title is None:
        if isinstance(obj, (Hit, dict)):
            title = obj.get("name")
        else:
            title = obj.title
        kw["title"] = title
    return kw


#
# Page level (all functions prefixed by 'page_').
#
class BasePageView:
    Model = WikiPage
    pk = "title"
    Form = WikiPageForm
    base_template = "wiki/_base.html"
    view_endpoint = ""

    @property
    def is_home_page(self):
        return self.obj is not None and self.obj.title == "Home"

    @property
    def template_kwargs(self) -> Dict[str, Union[WikiPageForm, WikiPage]]:
        """Template render arguments.

        You can override `base_template` for instance. Only `view` and
        `form` cannot be overriden.
        """
        kw = super().template_kwargs
        kw["page"] = self.obj
        kw["viewers"] = object_viewers(self.obj)
        return kw

    def init_object(
        self, args: Tuple[Any], kwargs: Dict[str, Any]
    ) -> Tuple[Tuple[Any], Dict[str, Any]]:
        title = kwargs["title"] = request.args["title"].strip()
        if title:
            try:
                self.obj = get_page_by_title(title)
            except NoResultFound:
                if title == "Home":
                    self.obj = create_home_page()
                else:
                    flash(
                        _("This page doesn't exit. You must create it first."),
                        "warning",
                    )
                    url = url_for(
                        ".page_new", title=title, community_id=g.community.slug
                    )
                    self.redirect(url)
            actions.context["object"] = self.obj
        return args, kwargs

    def index_url(self):
        return url_for(".index", community_id=g.community.slug)

    def view_url(self) -> str:
        return url_for(
            self.view_endpoint, community_id=g.community.slug, title=self.obj.title
        )


class PageView(BasePageView, ObjectView):
    template = "wiki/page.html"
    view_endpoint = ".page"
    decorators = [
        default_view(wiki, WikiPage, id_attr=None, kw_func=wiki_page_default_view_kw)
    ]

    def init_object(
        self, args: Tuple[Any], kwargs: Dict[str, Any]
    ) -> Tuple[Tuple[Any], Dict[str, Any]]:
        args, kwargs = BasePageView.init_object(self, args, kwargs)
        if not self.obj:
            title = kwargs["title"]
            if title == "Home":
                self.obj = create_home_page()
            else:
                url = url_for(".page_edit", title=title, community_id=g.community.slug)
                return redirect(url)

        actions.context["object"] = self.obj
        viewtracker.record_hit(entity=self.obj, user=current_user)
        return args, kwargs


route("/page/")(PageView.as_view("page"))


class PageViewers(PageView):

    template = "wiki/page_readers.html"


route("/viewers")(PageViewers.as_view("page_viewers"))


class PageEdit(BasePageView, ObjectEdit):
    # template = 'wiki/page_edit.html'
    title = _("Edit page")
    last_revision: int = None
    _message_success = _l("Wiki page successfully edited.")

    def init_object(
        self, args: Tuple[Any], kwargs: Dict[str, Any]
    ) -> Tuple[Tuple, Dict[str, Any]]:
        if request.method != "POST":
            return super().init_object(args, kwargs)

        page_id = request.form.get("page_id") or None
        if page_id is not None:
            page_id = int(page_id)
            self.obj = WikiPage.query.get_or_404(page_id)
            actions.context["object"] = self.obj

        return args, kwargs

    def get_form_kwargs(self) -> Dict[str, Union[WikiPage, Permission, int]]:
        kwargs = ObjectEdit.get_form_kwargs(self)
        kwargs["page_id"] = kwargs["last_revision_id"] = None
        if self.obj is not None and self.obj.id:
            # if no 'id', then it's a new object (PageCreate). We shouldn't call
            # obj.last_revision, since it will issue a flush (thus creating obj and
            # the 1st revision, resulting in an "edit conflict")
            kwargs["page_id"] = self.obj.id
            kwargs["last_revision_id"] = self.obj.last_revision.id
        return kwargs

    def prepare_args(
        self, args: Tuple[Any], kwargs: Dict[str, Any]
    ) -> Tuple[Tuple, Dict[str, Any]]:
        super().prepare_args(args, kwargs)
        last_revision_id = self.form.last_revision_id.data
        if last_revision_id:
            self.last_revision = WikiPageRevision.query.filter(
                WikiPageRevision.page == self.obj,
                WikiPageRevision.id == last_revision_id,
            ).first()
        return args, kwargs

    def before_populate_obj(self) -> None:
        self.redirect_if_no_change()
        form = self.form
        self.obj.create_revision(form.body_src.data, form.message.data)

    # FIXME: does nothing !
    def redirect_if_no_change(self) -> Optional[Response]:
        form = self.form
        if all(f.data == f.object_data for f in (form.title, form.body_src)):
            flash(_("You didn't make any change to this page."))
            return self.redirect(url_for(self.obj))
        return None

    @property
    def activity_target(self) -> Union[CommunityPresenter, Community]:
        return self.obj.community

    def form_invalid(self):
        is_conflict = self.form.errors.pop("last_revision_id", None)
        if not self.form.errors and is_conflict is not None:
            # no other error on form, just edit conflict: show new text
            self.form.last_revision_id.errors = []
            field = self.form.body_src
            current = self.obj.last_revision
            self.form.last_revision_id.data = current.id  # update edited revision
            self.redirect_if_no_change()  # same edition? don't bother

            if self.last_revision.body_src == current.body_src and self.form.validate():
                # only title change? cannot show diff: save if valid
                return self.form_valid()
            else:
                edited_src = field.data
                field.data = current.body_src
                edited_diff = [
                    l
                    for l in difflib.ndiff(
                        self.last_revision.body_src.splitlines(True),
                        edited_src.splitlines(True),
                    )
                    if not l[0] == "?"
                ]
                current_diff = [
                    l
                    for l in difflib.ndiff(
                        self.last_revision.body_src.splitlines(True),
                        current.body_src.splitlines(True),
                    )
                    if not l[0] == "?"
                ]
                ctx = {
                    "current": current,
                    "current_diff": current_diff,
                    "edited_diff": edited_diff,
                }
                field.errors.append(
                    Markup(render_template("wiki/edit_conflict_error.html", **ctx))
                )

        return None


route("/edit")(PageEdit.as_view("page_edit", view_endpoint=".page"))


class PageCreate(PageEdit, ObjectCreate):
    title = _l("Create page")
    _message_success = _l("Wiki page successfully created.")

    get_form_kwargs = ObjectCreate.get_form_kwargs

    def init_object(
        self, args: Tuple[Any], kwargs: Dict[str, Any]
    ) -> Tuple[Tuple[Any], Dict[str, Any]]:
        args, kwargs = ObjectCreate.init_object(self, args, kwargs)
        self.obj.community = g.community
        session = sa.orm.object_session(self.obj)
        if session:
            sa.orm.session.make_transient(self.obj)
            for rev in self.obj.revisions:
                sa.orm.session.make_transient(rev)
        return args, kwargs


route("/new")(PageCreate.as_view("page_new", view_endpoint=".page"))


@route("/source/")
def page_source() -> Union[str, Response]:
    title = request.args["title"].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        return redirect(
            url_for(".page_edit", title=title, community_id=g.community.slug)
        )

    actions.context["object"] = page
    return render_template("wiki/source.html", page=page)


@route("/changes/")
def page_changes() -> Union[str, Response]:
    title = request.args["title"].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        url = url_for(".page_edit", title=title, community_id=g.community.slug)
        return redirect(url)
    revisions = page.revisions
    revisions = sorted(revisions, key=lambda x: -x.number)
    actions.context["object"] = page
    return render_template("wiki/changes.html", page=page, revisions=revisions)


@route("/compare/")
def page_compare() -> Union[str, Response]:
    title = request.args["title"].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        return redirect(
            url_for(".page_edit", title=title, community_id=g.community.slug)
        )
    revisions = page.revisions
    revisions = sorted(revisions, key=lambda x: x.number)
    revs_to_compare = []
    for arg in request.args:
        if arg.startswith("rev"):
            revs_to_compare.append(int(arg[3:]))
    if len(revs_to_compare) != 2:
        flash(_("You must check exactly 2 revisions."), "error")
        url = url_for(".page_changes", title=title, community_id=g.community.slug)
        return redirect(url)

    revs_to_compare.sort()
    from_rev = revisions[revs_to_compare[0]]
    to_rev = revisions[revs_to_compare[1]]
    assert from_rev.number == revs_to_compare[0]
    assert to_rev.number == revs_to_compare[1]

    from_lines = from_rev.body_src.splitlines(1)
    to_lines = to_rev.body_src.splitlines(1)

    differ = difflib.Differ(charjunk=difflib.IS_CHARACTER_JUNK)
    diff = differ.compare(from_lines, to_lines)
    diff_lines = [line for line in diff if not line.startswith("?")]

    actions.context["object"] = page
    ctx = {"page": page, "diff": diff_lines, "rev1": from_rev, "rev2": to_rev}
    return render_template("wiki/compare.html", **ctx)


@route("/delete/", methods=["POST"])
def page_delete() -> Response:
    title = request.form["title"].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        flash(_("This page doesn't exist"), "error")
        return redirect(url_for(".index", community_id=g.community.slug))

    db.session.delete(page)

    app = unwrap(current_app)
    community = g.community._model
    activity.send(app, actor=current_user, verb="delete", object=page, target=community)

    db.session.commit()
    flash(_("Page %(title)s deleted.", title=title))
    return redirect(url_for(".index", community_id=g.community.slug))


@route("/attachments")
def attachment_download() -> Response:
    title = request.args["title"].strip()
    attachment_id = int(request.args["attachment"])
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        raise NotFound()

    attachment = WikiPageAttachment.query.get(attachment_id)
    assert attachment is not None
    assert attachment.wikipage is page

    response = make_response(attachment.content)
    response.headers["content-length"] = attachment.content_length
    response.headers["content-type"] = attachment.content_type
    filename = quote(attachment.name.encode("utf8"))
    content_disposition = f'attachment;filename="{filename}"'
    response.headers["content-disposition"] = content_disposition
    return response


@route("/attachments", methods=["POST"])
@csrf.protect
def attachment_upload() -> Response:
    title = request.args["title"].strip()
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        raise NotFound()

    files = request.files.getlist("attachments")
    saved_count = 0

    for f in files:
        name = f.filename
        if not isinstance(name, str):
            name = str(f.filename, encoding="utf-8", errors="ignore")

        # FIXME: do something instead of just skipping the attachement
        if not name:
            continue

        attachment = WikiPageAttachment(name=name)
        attachment.wikipage = page
        attachment.set_content(f.read(), f.content_type)
        db.session.add(attachment)
        saved_count += 1

    if saved_count:
        db.session.commit()
        flash(
            _n(
                "One new document successfully uploaded",
                "%(num)d new documents successfully uploaded",
                count=saved_count,
                num=len(files),
            ),
            "success",
        )
    else:
        flash(_("No file uploaded."))

    return redirect(url_for(page))


@route("/attachments/delete", methods=["POST"])
@csrf.protect
def attachment_delete():
    title = request.args["title"].strip()
    attachment_id = int(request.args["attachment"])
    try:
        page = get_page_by_title(title)
    except NoResultFound:
        raise NotFound()

    attachment = WikiPageAttachment.query.get(attachment_id)
    assert attachment is not None
    assert attachment.wikipage is page

    if request.form.get("action") == "delete":
        name = attachment.name
        db.session.delete(attachment)
        db.session.commit()
        flash(_('Attachment "{name}" has been deleted').format(name=name))

    return redirect(url_for(page))


#
# Wiki-level (prefixed by 'wiki_')
#
@route("/pages")
def wiki_pages():
    pages = (
        WikiPage.query.filter(WikiPage.community_id == g.community.id)
        .order_by(WikiPage.title)
        .all()
    )
    return render_template("wiki/all_pages.html", pages=pages)


@route("/help")
def wiki_help():
    src = open(Path(__file__).parent / "data" / "help.txt").read()
    body = Markup(markdown(src))
    return render_template("wiki/help.html", body=body)


@route("/export")
def wiki_export():
    # TODO
    return "Not done yet"


#
# Util
#
def get_page_by_title(title: str) -> WikiPage:
    title = title.strip()
    page = WikiPage.query.filter(
        WikiPage.community_id == g.community.id, WikiPage.title == title
    ).one()
    return page


def create_home_page() -> WikiPage:
    path = Path(__file__).parent / "data" / "default_page.txt"
    default_src = path.open("rt").read()
    page = WikiPage(title="Home", body_src=default_src)
    page.community_id = g.community.id
    db.session.add(page)
    db.session.commit()
    return page



import fileinput
import logging
import sys
from email.parser import FeedParser

import click
from flask.cli import with_appcontext

from .tasks import check_maildir, process_email

logger = logging.getLogger(__name__)


@click.command()
@with_appcontext
def inject_email(filename="-"):
    """Read one email from stdin, parse it, forward it in a celery task to be
    persisted."""
    _inject_email(filename)


def _inject_email(filename="-"):
    parser = FeedParser()

    if logger.level is logging.NOTSET:
        logger.setLevel(logging.INFO)

    try:
        # iterate over stdin
        for line in fileinput.input(filename):
            parser.feed(line)
    except KeyboardInterrupt:
        logger.info("Aborted by user, exiting.")
        sys.exit(1)
    except BaseException:
        logger.error("Error during email parsing", exc_info=True)
        sys.exit(1)
    finally:
        # close the parser to generate a email.message
        message = parser.close()
        fileinput.close()

    if message:
        # make sure no email.errors are present
        if not message.defects:
            process_email.delay(message)
        else:
            logger.error(
                "email has defects, message content:\n"
                "------ START -------\n"
                "%s"
                "\n------ END -------\n",
                message,
                extra={"stack": True},
            )
    else:
        logger.error("no email was parsed from stdin", extra={"stack": True})


@click.command()
@with_appcontext
def check_email():
    """Read one email from current user Maildir, parse it, forward it in a
    celery task to be persisted."""

    check_maildir()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['abilian',
 'abilian.sbe',
 'abilian.sbe.apps',
 'abilian.sbe.apps.calendar',
 'abilian.sbe.apps.calendar.tests',
 'abilian.sbe.apps.communities',
 'abilian.sbe.apps.communities.tests',
 'abilian.sbe.apps.communities.views',
 'abilian.sbe.apps.documents',
 'abilian.sbe.apps.documents.cmis',
 'abilian.sbe.apps.documents.tests',
 'abilian.sbe.apps.documents.views',
 'abilian.sbe.apps.documents.webdav',
 'abilian.sbe.apps.forum',
 'abilian.sbe.apps.forum.tests',
 'abilian.sbe.apps.main',
 'abilian.sbe.apps.notifications',
 'abilian.sbe.apps.notifications.tasks',
 'abilian.sbe.apps.notifications.views',
 'abilian.sbe.apps.preferences',
 'abilian.sbe.apps.preferences.panels',
 'abilian.sbe.apps.social',
 'abilian.sbe.apps.social.tests',
 'abilian.sbe.apps.social.views',
 'abilian.sbe.apps.wall',
 'abilian.sbe.apps.wiki',
 'abilian.sbe.apps.wiki.tests']

package_data = \
{'': ['*'],
 'abilian.sbe': ['static/csv/*',
                 'static/fileicons/*',
                 'static/icons/*',
                 'static/images/*',
                 'static/img/*',
                 'static/js/*',
                 'static/less/*',
                 'static/less/modules/*',
                 'static/moment/*',
                 'static/pdfjs/*',
                 'static/pdfjs/cmaps/*',
                 'static/pdfjs/images/*',
                 'static/pdfjs/locale/*',
                 'static/pdfjs/locale/en-US/*',
                 'static/pdfjs/locale/fr/*',
                 'static/vendor/*',
                 'templates/*',
                 'translations/*',
                 'translations/es/LC_MESSAGES/*',
                 'translations/fr/LC_MESSAGES/*',
                 'translations/tr/LC_MESSAGES/*',
                 'translations/zh/LC_MESSAGES/*'],
 'abilian.sbe.apps.calendar': ['templates/calendar/*'],
 'abilian.sbe.apps.communities': ['templates/community/*'],
 'abilian.sbe.apps.communities.views': ['data/*'],
 'abilian.sbe.apps.documents': ['templates/cmis/*', 'templates/documents/*'],
 'abilian.sbe.apps.documents.tests': ['data/*', 'data/dummy_files/*'],
 'abilian.sbe.apps.forum': ['templates/forum/*', 'templates/forum/mail/*'],
 'abilian.sbe.apps.forum.tests': ['data/*'],
 'abilian.sbe.apps.notifications': ['templates/notifications/*'],
 'abilian.sbe.apps.preferences': ['templates/preferences/*'],
 'abilian.sbe.apps.social': ['templates/social/*', 'templates/social/mail/*'],
 'abilian.sbe.apps.wall': ['templates/wall/*'],
 'abilian.sbe.apps.wiki': ['data/*', 'templates/wiki/*']}

install_requires = \
['abilian-core>=0.11,<0.12',
 'chardet',
 'flask-babel<2',
 'html2text>=2020.1.16,<2021.0.0',
 'itsdangerous<2',
 'jinja2<3',
 'langid>=1.1',
 'markdown>=3,<4',
 'openpyxl>=2.3.0,<3.0.0',
 'toolz',
 'validate_email',
 'wtforms<2.2',
 'xlwt']

setup_kwargs = {
    'name': 'abilian-sbe',
    'version': '0.5.21',
    'description': 'Social Business platform (entreprise collaboration and information management)',
    'long_description': "About\n=====\n\nAbilian SBE (Social Business Engine) is a platform for social business applications, and more specifically collaborative / enterprise 2.0 business applications, such as enterprise social networks (ESN).\n\nAbilian SBE revolves around the concept of *communities*, which are collaborative spaces with services such as lightweight document management, discussions, wikis, user timelines, etc.\n\nIt is based on the `Abilian Core <http://abilian-core.readthedocs.org/en/latest/>`_ project which provide the basic services, on top of Flask and SQLAlchemy.\n\nAbilian SBE is used by several major customers in production, since mid 2013.\n\n\n\n.. image:: https://circleci.com/gh/abilian/abilian-sbe.svg?style=svg\n   :target: https://circleci.com/gh/abilian/abilian-sbe\n\n.. image:: https://static.pepy.tech/badge/abilian-core\n   :target: https://pepy.tech/project/abilian-core\n\n\nScreenshots\n-----------\n\n.. image:: https://raw.githubusercontent.com/abilian/abilian-sbe/master/docs/images/screenshot-3.png\n\n.. image:: https://raw.githubusercontent.com/abilian/abilian-sbe/master/docs/images/screenshot-2.png\n\n.. image:: https://raw.githubusercontent.com/abilian/abilian-sbe/master/docs/images/screenshot-1.png\n\n\nInstall\n=======\n\nPrerequisites (native dependencies)\n-----------------------------------\n\n- Python 2.7, ``virtualenv``, ``pip``\n- `Redis <http://redis.io/>`_\n- Sqlite, or a postgresql database.\n- A few image manipulation libraries (``libpng``, ``libjpeg``...)\n- ``poppler-utils``, ``unoconv``, ``LibreOffice``, ``ImageMagick``.\n- `{Less} <http://lesscss.org/>`__ css pre-processor\n- A Java environment (JRE 1.7 for example). The `closure compiler\n  <https://developers.google.com/closure/compiler/>`_ is used for minifying\n  javascript files. You don't have to install the compiler yourself, but a Java\n  environment is required.\n\nGet a working application\n-------------------------\n\nThe following commands will create a virtualenv for the application,\ninstall a script named ``abilian_sbe``, launch development server and\nopen a setupwizard in your browser:\n\n.. code:: bash\n\n    $ virtualenv sbe\n    $ cd sbe; source bin/activate\n    $ pip install -U setuptools pip\n    $ pip install abilian-sbe\n    $ python -m abilian.sbe.app setup_sbe_app\n\nMAC OS + Homebrew\n-----------------\n\nYou will need to install the following packages using homebrew\n(**before** running ``pip install ...``):\n\n::\n\n    brew install python2.7 jpeg git libmagic poppler imagemagick\n\nRunning with gunicorn\n---------------------\n\n.. code:: bash\n\n    gunicorn 'abilian.sbe.app.create_app()'\n\n\nDeveloping\n==========\n\nBranching policy\n----------------\n\nWe currently have two branches:\n\n- `master` for stable version (0.3.x)\n- `develop` for next version (0.4.devx)\n\nTesting\n-------\n\nShort test\n~~~~~~~~~~\n\nMake sure all the dependencies are installed (cf. above), then run ``make\ntest``.\n\nWith coverage\n~~~~~~~~~~~~~\n\nRun ``make test-with-coverage``.\n\nFull test suite\n~~~~~~~~~~~~~~~\n\nInstall `tox <http://pypi.python.org/pypi/tox>`_. Run ``tox -e ALL``.\n\n2 environments are available:\n\n- ``py27``: uses in-memory sqlite\n- ``py27_postgres``: uses local postgresql server (you need to first create a\n   database, and user/password; tox uses environment variables\n   ``POSTGRES_HOST``, ``POSTGRES_PORT``, ``POSTGRES_DB``, ``POSTGRES_USER``,\n   ``POSTGRES_PASSWORD``)\n\nBuild Status\n------------\n\nThe project is under continuous integration with Travis and CircleCI:\n\n.. image:: https://travis-ci.org/abilian/abilian-sbe.svg?branch=master\n   :target: https://travis-ci.org/abilian/abilian-sbe\n\n.. image:: https://coveralls.io/repos/abilian/abilian-sbe/badge.svg?branch=master\n   :target: https://coveralls.io/r/abilian/abilian-sbe?branch=master\n\nLinks\n=====\n\n- `Discussion list (Google Groups) <https://groups.google.com/forum/#!foru      m/abilian-users>`_\n- `Documentation <http://docs.abilian.com/>`_\n- `GitHub repository <https://github.com/abilian/abilian-sbe>`_\n- `Corporate support <http://www.abilian.com>`_\n",
    'author': 'Abilian SAS',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abilian/abilian-sbe',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Any, Dict

from flask import g
from flask import url_for as url_for_orig
from flask.blueprints import BlueprintSetupState
from flask_login import current_user

from abilian.i18n import _l
from abilian.sbe.apps.communities.security import is_manager
from abilian.services.security import MANAGE, WRITE, security
from abilian.web.action import Action, FAIcon, ModalActionMixin, actions

from .repository import repository


def url_for(endpoint, **kw):
    return url_for_orig(endpoint, community_id=g.community.slug, **kw)


class CmisContentAction(Action):
    sbe_type: str = ""
    permission: str = ""

    def __init__(self, *args, **kwargs) -> None:
        if "permission" in kwargs:
            self.permission = kwargs.pop("permission")

        Action.__init__(self, *args, **kwargs)

    def pre_condition(self, ctx: Dict[str, Any]) -> bool:
        obj = ctx["object"]
        ok = obj.sbe_type == self.sbe_type

        if ok and self.permission is not None:
            ok = self.has_access(self.permission, obj)

        return ok

    def has_access(self, permission: str, obj: Any) -> bool:
        return repository.has_permission(current_user, permission, obj)


class BaseFolderAction(CmisContentAction):
    """Apply to all folders, including root folder."""

    sbe_type = "cmis:folder"


class FolderButtonAction(BaseFolderAction):

    _std_template_string = (
        '<button class="btn {{ action.css_class }}" name="action" '
        'value="{{ action.name }}" title="{{ action.title }}">'
        "{{ action.icon }}</button>"
    )

    _modal_template_string = (
        '<a class="btn {{ action.css_class }}" href="{{ url }}" '
        'data-toggle="modal" role="button" title="{{ action.title }}">'
        "{{ action.icon }}</a>"
    )

    def __init__(self, *args, **kwargs) -> None:
        self.modal = False

        if "modal" in kwargs:
            self.modal = kwargs.pop("modal")

        css_class = kwargs.pop("css_class", "btn-default")
        self.CSS_CLASS = f"{self.CSS_CLASS} {css_class}"

        BaseFolderAction.__init__(self, *args, **kwargs)

    @property
    def template_string(self):
        return self._modal_template_string if self.modal else self._std_template_string


class FolderAction(BaseFolderAction):
    """Apply to all folders except root folder."""

    sbe_type = "cmis:folder"

    def pre_condition(self, ctx):
        return (
            super().pre_condition(ctx) and ctx["object"] is not repository.root_folder
        )


class FolderPermisionsAction(BaseFolderAction):
    """Apply to all folders except root folder."""

    sbe_type = "cmis:folder"

    def pre_condition(self, ctx):
        return (
            super(BaseFolderAction, self).pre_condition(ctx) and ctx["object"].depth > 1
        )


class FolderModalAction(ModalActionMixin, FolderAction):
    pass


class DocumentAction(CmisContentAction):
    sbe_type = "cmis:document"


class DocumentModalAction(ModalActionMixin, DocumentAction):
    pass


class RootFolderAction(CmisContentAction):
    """Apply only for root folder."""

    def pre_condition(self, ctx):
        return ctx["object"] is repository.root_folder


_checkin_template_action = """
<form method="POST" action="{{ url }}" encoding="multipart/form-data" target="_new">
  {{ csrf.field() }}
  <button type="submit" class="btn btn-link" name="action"
          value="{{ action.name}}"
     onclick="window.setTimeout(function () { window.location.reload() }, 500)">
    {%- if action.icon %}{{ action.icon }} {% endif %}
    {{ action.title }}
  </button>
</form>
"""

_lock_template_action = """
<form method="POST" action="{{ url }}" encoding="multipart/form-data">
  {{ csrf.field() }}
  <button type="submit" class="btn btn-link" name="action"
          value="{{ action.name}}">
    {%- if action.icon %}{{ action.icon }} {% endif %}
    {{ action.title }}
  </button>
</form>
"""

_actions = (
    # Folder listing action buttons ##########
    FolderButtonAction(
        "documents:folder-listing", "download", _l("Download"), icon="download"
    ),
    FolderButtonAction(
        "documents:folder-listing",
        "move-files",
        _l("Move to another folder"),
        icon="move",
        url="#modal-move-files",
        modal=True,
        permission=WRITE,
    ),
    FolderButtonAction(
        "documents:folder-listing",
        "delete",
        _l("Delete"),
        permission=WRITE,
        icon="trash",
        css_class="btn-danger",
    ),
    FolderButtonAction(
        "documents:folder-listing",
        "change-owner",
        _l("Change owner"),
        icon="user",
        url="#modal-change-owner",
        modal=True,
        permission=MANAGE,
    ),
    # Folder left bar actions ##########
    # view
    RootFolderAction(
        "documents:content",
        "view",
        _l("List content"),
        icon="list",
        condition=(lambda ctx: security.has_role(current_user, "admin")),
        url=lambda ctx: url_for(".folder_view", folder_id=ctx["object"].id),
    ),
    # view
    FolderAction(
        "documents:content",
        "view",
        _l("List content"),
        icon="list",
        url=lambda ctx: url_for(".folder_view", folder_id=ctx["object"].id),
    ),
    # Descendants
    FolderAction(
        "documents:content",
        "descendants",
        _l("View descendants"),
        icon=FAIcon("code-fork fa-rotate-90"),
        url=lambda ctx: url_for(".descendants_view", folder_id=ctx["object"].id),
    ),
    # upload
    FolderModalAction(
        "documents:content",
        "upload_files",
        _l("Upload file(s)"),
        icon="upload",
        url="#modal-upload-files",
        permission=WRITE,
    ),
    # edit
    FolderModalAction(
        "documents:content",
        "edit",
        _l("Edit properties"),
        icon="pencil",
        url="#modal-edit",
        permission=WRITE,
    ),
    # new folder
    FolderModalAction(
        "documents:content",
        "new_folder",
        _l("New folder"),
        icon="plus",
        url="#modal-new-folder",
        permission=WRITE,
    ),
    # # members
    # FolderAction(
    #   'documents:content', 'members', _l('Members'), icon='user',
    #   url=lambda ctx: url_for(".members", folder_id=ctx['object'].id)),
    # permissions
    FolderPermisionsAction(
        "documents:content",
        "permissions",
        _l("Permissions"),
        icon="lock",
        url=lambda ctx: url_for(".permissions", folder_id=ctx["object"].id),
        permission=MANAGE,
    ),
    # Document actions ##########
    # View / preview in browser
    DocumentAction(
        "documents:content",
        "preview",
        _l("View in browser"),
        icon="eye-open",
        url=lambda ctx: url_for(".document_preview", doc_id=ctx["object"].id),
        condition=(
            lambda ctx: ctx["object"].antivirus_ok
            and ctx["object"].content_type
            in ("text/html", "text/plain", "application/pdf")
        ),
    ),
    # viewers
    DocumentModalAction(
        "documents:content",
        "document_viewers",
        _l("Viewers list"),
        icon="user",
        condition=lambda ctx: is_manager(context=ctx),
        url=lambda ctx: url_for(".document_viewers", doc_id=ctx["object"].id),
    ),
    # edit
    DocumentModalAction(
        "documents:content",
        "edit",
        _l("Edit properties"),
        icon="pencil",
        url="#modal-edit",
        permission=WRITE,
    ),
    # Checkin / Checkout
    DocumentAction(
        "documents:content",
        "checkout",
        _l("Checkout (Download for edit)"),
        icon="download",
        url=lambda ctx: url_for(".checkin_checkout", doc_id=ctx["object"].id),
        condition=lambda ctx: ctx["object"].lock is None,
        template_string=_checkin_template_action,
    ),
    # DocumentAction(
    #   'documents:content', 'lock', _l(u'Lock for edit'),
    #   icon='lock',
    #   url=lambda ctx: url_for('.checkin_checkout', doc_id=ctx['object'].id),
    #   condition=lambda ctx: ctx['object'].lock is None,
    #   template_string=_lock_template_action,
    # ),
    DocumentAction(
        "documents:content",
        "unlock",
        _l("Unlock"),
        icon=FAIcon("unlock"),
        url=lambda ctx: url_for(".checkin_checkout", doc_id=ctx["object"].id),
        condition=lambda ctx: ctx["object"].lock is not None,
        template_string=_lock_template_action,
    ),
    # upload-new / checkin
    DocumentModalAction(
        "documents:content",
        "upload",
        _l("Upload new version"),
        icon="upload",
        url="#modal-upload-new-version",
        # either not locked, either user is owner
        condition=lambda ctx: not ctx["object"].lock or ctx["object"].lock.is_owner(),
        permission=WRITE,
    ),
    # send by email
    DocumentModalAction(
        "documents:content",
        "send_by_email",
        _l("Send by email"),
        icon="envelope",
        url="#modal-send-by-email",
        condition=lambda ctx: ctx["object"].antivirus_ok,
        permission=WRITE,
    ),
    # delete
    DocumentModalAction(
        "documents:content",
        "delete",
        _l("Delete"),
        icon="trash",
        url="#modal-delete",
        permission=WRITE,
    ),
    # refresh preview
    DocumentAction(
        "documents:content",
        "refresh_preview",
        _l("Refresh preview"),
        icon="refresh",
        url=lambda ctx: url_for(".refresh_preview", doc_id=ctx["object"].id),
        condition=lambda ctx: ctx["object"].antivirus_ok,
        permission=MANAGE,
    ),
)


def register_actions(state: BlueprintSetupState) -> None:
    if not actions.installed(state.app):
        return

    with state.app.app_context():
        actions.register(*_actions)



from flask import Markup, render_template


class UserPhotoInputWidget:
    def __call__(self, field, **kwargs):
        kwargs.setdefault("id", field.id)
        user_id = None

        if "widget_options" in kwargs:
            options = kwargs.pop("widget_options")
            user_id = options.get("user_id")

        return Markup(
            render_template(
                "social/widget_user_photo.html",
                field=field,
                attrs=kwargs,
                user_id=user_id,
            )
        )



from io import BytesIO
from pathlib import Path
from typing import IO
from zipfile import ZipFile

import flask_mail
import pytest
from flask import g, get_flashed_messages
from flask.ctx import RequestContext
from flask.testing import FlaskClient
from pytest import fixture
from toolz import first
from werkzeug.datastructures import FileStorage

from abilian.core.sqlalchemy import SQLAlchemy
from abilian.sbe.app import Application
from abilian.sbe.apps.communities.models import WRITER, Community
from abilian.sbe.apps.communities.presenters import CommunityPresenter
from abilian.sbe.apps.documents.models import Folder
from abilian.sbe.apps.documents.views import util as view_util
from abilian.testing.util import client_login, path_from_url
from abilian.web.util import url_for


def open_file(filename: str) -> IO[bytes]:
    path = Path(__file__).parent / "data" / "dummy_files" / filename
    return path.open("rb")


def uid_from_url(url):
    return int(url.split("/")[-1])


@fixture
def community(community1: Community, db: SQLAlchemy) -> Community:
    community = community1
    user = community.test_user
    root_folder = Folder(title="root")
    db.session.add(root_folder)
    db.session.flush()

    community.type = "participative"
    community.set_membership(user, WRITER)
    db.session.commit()

    return community


def test_util_create(
    app: Application,
    client: FlaskClient,
    db: SQLAlchemy,
    community: Community,
    req_ctx: RequestContext,
) -> None:
    folder = community.folder
    user = community.test_user

    with client_login(client, user):
        g.community = CommunityPresenter(community)
        name = "document"
        fs = FileStorage(BytesIO(b"content"), filename=name, content_type="text/plain")

        doc = view_util.create_document(folder, fs)
        db.session.flush()
        assert doc.parent == folder
        assert doc.name == name

        # test upload with same name: should be renamed
        fs = FileStorage(BytesIO(b"content"), filename=name, content_type="text/plain")
        doc2 = view_util.create_document(folder, fs)
        db.session.flush()
        assert doc2.parent == folder
        assert len(folder.children) == 2
        assert doc2.name == f"{name}-1"

        messages = get_flashed_messages()
        assert len(messages) == 1


def test_home(
    app: Application,
    client: FlaskClient,
    db: SQLAlchemy,
    community: Community,
    req_ctx: RequestContext,
) -> None:
    folder = community.folder
    user = community.test_user

    with client_login(client, user):
        response = client.get(url_for("documents.index", community_id=community.slug))
        assert response.status_code == 302
        path = path_from_url(response.location)
        expected = f"/communities/{community.slug}/docs/folder/{folder.id}"
        assert path == expected


def _test_upload(
    community: Community,
    client: FlaskClient,
    title: str,
    content_type: str,
    test_preview: bool = True,
    assert_preview_available: bool = True,
) -> None:
    data = {"file": (open_file(title), title, content_type), "action": "upload"}

    folder = community.folder
    url = url_for(
        "documents.folder_post", community_id=community.slug, folder_id=folder.id
    )
    response = client.post(url, data=data)
    assert response.status_code == 302

    doc = folder.children[0]
    assert doc.title == title

    url = url_for("documents.document_view", community_id=community.slug, doc_id=doc.id)
    response = client.get(url)
    assert response.status_code == 200

    url = url_for(
        "documents.document_download", community_id=community.slug, doc_id=doc.id
    )
    response = client.get(url)
    assert response.status_code == 200
    assert response.headers["Content-Type"] == content_type

    content = open_file(title).read()
    assert response.data == content

    if test_preview:
        url = url_for(
            "documents.document_preview_image",
            community_id=community.slug,
            doc_id=doc.id,
            size=500,
        )
        response = client.get(url)
        if assert_preview_available:
            assert response.status_code == 200
            assert response.headers["Content-Type"] == "image/jpeg"
        else:
            # redirect to 'missing image'
            assert response.status_code == 302
            assert response.headers["Cache-Control"] == "no-cache"

    url = url_for(
        "documents.document_delete", community_id=community.slug, doc_id=doc.id
    )
    response = client.post(url)
    assert response.status_code == 302

    url = url_for("documents.document_view", community_id=community.slug, doc_id=doc.id)
    response = client.get(url)
    assert response.status_code == 404


def test_text_upload(
    client: FlaskClient, community: Community, req_ctx: RequestContext
) -> None:
    name = "wikipedia-fr.txt"
    user = community.test_user
    with client_login(client, user):
        _test_upload(community, client, name, "text/plain", test_preview=False)


def test_pdf_upload(
    client: FlaskClient, community: Community, req_ctx: RequestContext
) -> None:
    name = "onepage.pdf"
    user = community.test_user
    with client_login(client, user):
        _test_upload(community, client, name, "application/pdf")


def test_image_upload(
    client: FlaskClient, community: Community, req_ctx: RequestContext
) -> None:
    name = "picture.jpg"
    user = community.test_user
    with client_login(client, user):
        _test_upload(community, client, name, "image/jpeg")


@pytest.mark.skip()  # FIXME: magic detection mismatch?
def test_binary_upload(client, community, req_ctx):
    name = "random.bin"
    user = community.test_user
    with client_login(client, user):
        _test_upload(
            community,
            client,
            name,
            "application/octet-stream",
            assert_preview_available=False,
        )


def test_zip_upload_uncompress(
    community: Community, db: SQLAlchemy, client: FlaskClient, req_ctx: RequestContext
) -> None:
    subfolder = Folder(title="folder 1", parent=community.folder)
    db.session.add(subfolder)
    db.session.flush()

    folder = community.folder
    files = [
        (BytesIO(b"A document"), "existing-doc", "text/plain"),
        (open_file("content.zip"), "content.zip", "application/zip"),
    ]
    data = {"file": files, "action": "upload", "uncompress_files": True}
    url = url_for(
        "documents.folder_post", community_id=community.slug, folder_id=folder.id
    )
    user = community.test_user
    with client_login(client, user):
        response = client.post(url, data=data)

    assert response.status_code == 302
    expected = {"existing-doc", "folder 1", "existing-doc-1"}
    assert expected == {f.title for f in folder.children}
    expected = {"folder 1", "existing-doc-1"}
    assert expected == {f.title for f in folder.subfolders}


def test_zip(
    community: Community, client: FlaskClient, req_ctx: RequestContext
) -> None:
    user = community.test_user
    with client_login(client, user):
        title = "onepage.pdf"
        content_type = "application/pdf"
        data = {"file": (open_file(title), title, content_type), "action": "upload"}
        folder = community.folder
        url = url_for(
            "documents.folder_post", community_id=community.slug, folder_id=folder.id
        )
        response = client.post(url, data=data)
        assert response.status_code == 302

        doc = folder.children[0]
        data = {"action": "download", "object-selected": [f"cmis:document:{doc.id:d}"]}
        url = url_for(
            "documents.folder_post", community_id=community.slug, folder_id=folder.id
        )
        response = client.post(url, data=data)
        assert response.status_code == 200
        assert response.content_type == "application/zip"

        zipfile = ZipFile(BytesIO(response.data))
        assert [zipfile.namelist()[0]] == [title]


def test_recursive_zip(
    community: Community, client: FlaskClient, req_ctx: RequestContext
) -> None:
    user = community.test_user
    with client_login(client, user):
        data1 = {"action": "new", "title": "my folder"}
        folder = community.folder
        url = url_for(
            "documents.folder_post", community_id=community.slug, folder_id=folder.id
        )
        response = client.post(url, data=data1)
        assert response.status_code == 302

        my_folder = folder.children[0]

        title = "onepage.pdf"
        content_type = "application/pdf"
        data2 = {"file": (open_file(title), title, content_type), "action": "upload"}
        url = url_for(
            "documents.folder_post", community_id=community.slug, folder_id=my_folder.id
        )
        response = client.post(url, data=data2)
        assert response.status_code == 302

        data3 = {
            "action": "download",
            "object-selected": [f"cmis:folder:{my_folder.id:d}"],
        }
        url = url_for(
            "documents.folder_post", community_id=community.slug, folder_id=folder.id
        )
        response = client.post(url, data=data3)
        assert response.status_code == 200
        assert response.content_type == "application/zip"

        zipfile = ZipFile(BytesIO(response.data))
        assert zipfile.namelist() == [f"my folder/{title}"]


def test_document_send_by_mail(
    app: Application, community: Community, client: FlaskClient, req_ctx: RequestContext
) -> None:
    mail = app.extensions["mail"]
    folder = community.folder
    user = community.test_user
    with client_login(client, user):
        # upload files
        for filename in ("ascii title.txt", "utf-8 est arrivé!.txt"):
            content_type = "text/plain"
            data = {
                "file": (BytesIO(b"file content"), filename, content_type),
                "action": "upload",
            }
            url = url_for(
                "documents.folder_post",
                community_id=community.slug,
                folder_id=folder.id,
            )
            client.post(url, data=data)

        ascii_doc = folder.children[0]
        unicode_doc = folder.children[1]

        def get_send_url(doc_id: int) -> str:
            return url_for(
                "documents.document_send", community_id=community.slug, doc_id=doc_id
            )

        # mail ascii filename
        with mail.record_messages() as outbox:
            url = get_send_url(ascii_doc.id)
            response = client.post(
                url,
                data={"recipient": "dest@example.com", "message": "Voilà un fichier"},
            )
            assert response.status_code == 302
            assert len(outbox) == 1

            msg = outbox[0]
            assert msg.subject == "[Abilian Test] Unknown sent you a file"
            assert msg.recipients == ["dest@example.com"]

            assert len(msg.attachments) == 1

            attachment = first(msg.attachments)
            assert isinstance(attachment, flask_mail.Attachment)
            assert attachment.filename == "ascii title.txt"

        # mail unicode filename
        with mail.record_messages() as outbox:
            url = get_send_url(unicode_doc.id)
            response = client.post(
                url,
                data={"recipient": "dest@example.com", "message": "Voilà un fichier"},
            )
            assert response.status_code == 302
            assert len(outbox) == 1

            msg = outbox[0]
            assert isinstance(msg, flask_mail.Message)
            assert msg.subject == "[Abilian Test] Unknown sent you a file"
            assert msg.recipients == ["dest@example.com"]
            assert len(msg.attachments) == 1

            attachment = first(msg.attachments)
            assert isinstance(attachment, flask_mail.Attachment)
            assert attachment.filename == "utf-8 est arrivé!.txt"
